using System;
using System.Collections.Generic;
using System.Reflection;
using System.Linq;
using Encompass.Exceptions;

namespace Encompass
{
    public class WorldBuilder
    {
        private readonly List<Engine> engines = new List<Engine>();
        private readonly DirectedGraph<Engine> engineGraph = new DirectedGraph<Engine>();

        private readonly ComponentManager componentManager;
        private readonly EntityManager entityManager;
        private readonly MessageManager messageManager;
        private readonly DrawLayerManager drawLayerManager;
        private readonly RenderManager renderManager;

        private readonly Dictionary<Type, HashSet<Engine>> messageTypeToEmitters = new Dictionary<Type, HashSet<Engine>>();
        private readonly Dictionary<Type, HashSet<Engine>> messageTypeToReaders = new Dictionary<Type, HashSet<Engine>>();

        public WorldBuilder()
        {
            var entitiesWithAddedComponents = new List<Guid>();
            var entitiesWithRemovedComponents = new List<Guid>();
            drawLayerManager = new DrawLayerManager();
            componentManager = new ComponentManager(drawLayerManager, entitiesWithAddedComponents, entitiesWithRemovedComponents);
            entityManager = new EntityManager(componentManager, entitiesWithAddedComponents, entitiesWithRemovedComponents);
            messageManager = new MessageManager();
            renderManager = new RenderManager(entityManager, componentManager, drawLayerManager);
        }

        public Entity CreateEntity()
        {
            return entityManager.CreateEntity();
        }

        public Engine AddEngine<TEngine>(TEngine engine) where TEngine : Engine
        {
            engine.AssignEntityManager(entityManager);
            engine.AssignComponentManager(componentManager);
            engine.AssignMessageManager(messageManager);

            engines.Add(engine);
            engineGraph.AddVertex(engine);

            if (engine is IEntityTracker)
            {
                entityManager.RegisterEntityTracker(engine as IEntityTracker);
            }

            var emitMessageAttribute = engine.GetType().GetCustomAttribute<Emits>(false);
            if (emitMessageAttribute != null)
            {
                foreach (var emitMessageType in engine.GetType().GetCustomAttribute<Emits>(false).emitMessageTypes)
                {
                    if (!messageTypeToEmitters.ContainsKey(emitMessageType))
                    {
                        messageTypeToEmitters.Add(emitMessageType, new HashSet<Engine>());
                    }

                    messageTypeToEmitters[emitMessageType].Add(engine);

                    if (messageTypeToReaders.ContainsKey(emitMessageType))
                    {
                        foreach (var reader in messageTypeToReaders[emitMessageType])
                        {
                            engineGraph.AddEdge(engine, reader);
                        }
                    }
                }
            }

            var readMessageAttribute = engine.GetType().GetCustomAttribute<Reads>(false);
            if (readMessageAttribute != null)
            {
                foreach (var readMessageType in engine.GetType().GetCustomAttribute<Reads>(false).readMessageTypes)
                {
                    if (!messageTypeToReaders.ContainsKey(readMessageType))
                    {
                        messageTypeToReaders.Add(readMessageType, new HashSet<Engine>());
                    }

                    messageTypeToReaders[readMessageType].Add(engine);

                    if (messageTypeToEmitters.ContainsKey(readMessageType))
                    {
                        foreach (var emitter in messageTypeToEmitters[readMessageType])
                        {
                            engineGraph.AddEdge(emitter, engine);
                        }
                    }
                }
            }

            return engine;
        }

        public TRenderer AddEntityRenderer<TRenderer>() where TRenderer : Renderer, new()
        {
            var renderer = new TRenderer();
            renderer.AssignEntityManager(entityManager);
            renderer.AssignComponentManager(componentManager);

            if (renderer is EntityRenderer)
            {
                entityManager.RegisterEntityTracker(renderer as IEntityTracker);
                renderManager.RegisterEntityRenderer(renderer as EntityRenderer);
            }

            return renderer;
        }

        public TRenderer AddGeneralRenderer<TRenderer>(int layer) where TRenderer : GeneralRenderer, new()
        {
            var renderer = new TRenderer();
            renderer.AssignEntityManager(entityManager);
            renderer.AssignComponentManager(componentManager);

            renderManager.RegisterGeneralRendererWithLayer(renderer, layer);

            return renderer;
        }

        public World Build()
        {
            if (engineGraph.Cyclic())
            {
                var cycles = engineGraph.SimpleCycles();
                var errorString = "Cycle(s) found in Engines: ";
                foreach (var cycle in cycles)
                {
                    errorString += "\n" +
                    string.Join(" -> ", cycle.Select((engine) => engine.GetType().Name)) +
                    " -> " +
                    cycle.First().GetType().Name;
                }
                throw new EngineCycleException(errorString);
            }

            var mutatedComponentTypes = new HashSet<Type>();
            var duplicateMutations = new List<Type>();
            var componentToEngines = new Dictionary<Type, List<Engine>>();

            foreach (var engine in engines)
            {
                var mutateAttribute = engine.GetType().GetCustomAttribute<Mutates>(false);
                if (mutateAttribute != null)
                {
                    foreach (var mutateComponentType in engine.GetType().GetCustomAttribute<Mutates>(false).mutateComponentTypes)
                    {
                        if (mutatedComponentTypes.Contains(mutateComponentType))
                        {
                            duplicateMutations.Add(mutateComponentType);
                        }
                        else
                        {
                            mutatedComponentTypes.Add(mutateComponentType);
                        }

                        if (!componentToEngines.ContainsKey(mutateComponentType))
                        {
                            componentToEngines[mutateComponentType] = new List<Engine>();
                        }

                        componentToEngines[mutateComponentType].Add(engine);
                    }
                }
            }

            if (duplicateMutations.Count > 0)
            {
                var errorString = "Multiple Engines mutate the same Component: ";
                foreach (var componentType in duplicateMutations)
                {
                    errorString += "\n" +
                    componentType.Name + " mutated by: " +
                    string.Join(", ", componentToEngines[componentType].Select((engine) => engine.GetType().Name));
                }

                throw new EngineMutationConflictException(errorString);
            }

            var engineOrder = new List<Engine>();
            foreach (var engine in engineGraph.TopologicalSort())
            {
                engineOrder.Add(engine);
            }

            var world = new World(
                engineOrder,
                entityManager,
                componentManager,
                messageManager,
                renderManager
            );

            componentManager.ActivateMarkedComponents();
            componentManager.DeactivateMarkedComponents();
            componentManager.RemoveMarkedComponents();

            entityManager.CheckEntitiesWithAddedComponents();
            entityManager.CheckEntitiesWithRemovedComponents();

            return world;
        }
    }
}
